from parallellinear.calculations.LinearCalculations import LinearCalculations
import pyopencl as cl
import numpy as np
import os


class ParallelLinear(LinearCalculations):


    def __init__(self, deferToNumpy = False, programs = None, ctx = None, queue = None, baseProgramNames=None):
        if programs is not None and ctx is not None and queue is not None and baseProgramNames is not None and type(ctx) == cl.Context and type(queue) == cl.CommandQueue:
            self.programsLoaded = True
            self.programs = programs
            self.ctx = ctx
            self.queue = queue
            self.baseProgramNames = baseProgramNames
        else:
            raise ValueError("ParalleLinear object not initialized properly.  Use ParallelLinear.loadPrograms(), or ParallelLinear.deferToNumpy() to initialize")
        self.deferToNumpy = deferToNumpy
        
    @classmethod
    def getLinearCalculator(cls, pyopencl_ctx=":2"):
        return cls._loadPrograms(cls, pyopencl_ctx)

    def _loadPrograms(cls, pyopencl_ctx=":2"):
        os.environ["PYOPENCL_CTX"] = pyopencl_ctx
        ctx = cl.create_some_context()
        queue = cl.CommandQueue(ctx)

        prg = cl.Program(ctx, """
        __kernel void testing(const unsigned int size, __global float* a_g, __global const float* b_g){
            unsigned int gid1 = get_global_id(0);
            unsigned int gid2 = get_global_id(1);
            a_g[gid1 + size * gid2] = a_g[gid1 + size * gid2] + b_g[gid1 + size * gid2];
        }
        __kernel void add(__global const float *a_g, __global const float *b_g, __global float *res_g){
            int gid = get_global_id(0);
            res_g[gid] = a_g[gid] + b_g[gid];
        }
        __kernel void sub(__global const float *a_g, __global const float *b_g, __global float *res_g){
            int gid = get_global_id(0);
            res_g[gid] = a_g[gid] - b_g[gid];
        }
        __kernel void element_wise_multiply(__global const float *a_g, __global const float *b_g, __global float *res_g){
            int gid = get_global_id(0);
            res_g[gid] = a_g[gid] * b_g[gid];
        }
        __kernel void add_scaler( __global float *a_g, const float scaler, __global float *res_g){
            int gid = get_global_id(0);
            res_g[gid] = a_g[gid] + scaler;
        }
        __kernel void sub_scaler( __global float *a_g, const float scaler, __global float *res_g){
            int gid = get_global_id(0);
            res_g[gid] = a_g[gid] - scaler;
        }
        __kernel void scaler_sub_from( __global float *a_g, const float scaler, __global float *res_g){
            int gid = get_global_id(0);
            res_g[gid] = scaler - a_g[gid];
        }
        __kernel void scale( __global const float *a_g, const float scaler, __global float *res_g){
            int gid = get_global_id(0);
            res_g[gid] = a_g[gid] * scaler;
        }
        __kernel void descale(__global const float *a_g, const float scaler, __global float *res_g){
            int gid = get_global_id(0);
            res_g[gid] = a_g[gid] / scaler;
        }

        __kernel void add_inplace(__global float *a_g, __global const float *b_g){
            int gid = get_global_id(0);
            a_g[gid] = a_g[gid] + b_g[gid];
        }
        __kernel void sub_inplace( __global float *a_g, __global const float *b_g){
            int gid = get_global_id(0);
            a_g[gid] = a_g[gid] - b_g[gid];
        }
        __kernel void element_wise_multiply_inplace(__global float *a_g, __global const float *b_g){
            int gid = get_global_id(0);
            a_g[gid] = a_g[gid] * b_g[gid];
        }
        __kernel void add_scaler_inplace( __global float *a_g, const float scaler){
            int gid = get_global_id(0);
            a_g[gid] = a_g[gid] + scaler;
        }
        __kernel void sub_scaler_inplace( __global float *a_g, const float scaler){
            int gid = get_global_id(0);
            a_g[gid] = a_g[gid] - scaler;
        }
        __kernel void scaler_sub_from_inplace( __global float *a_g, const float scaler){
            int gid = get_global_id(0);
            a_g[gid] = scaler - a_g[gid];
        }
        __kernel void scale_inplace( __global float *a_g, const float scaler){
            int gid = get_global_id(0);
            a_g[gid] = a_g[gid] * scaler;
        }
        __kernel void descale_inplace(__global float *a_g, const float scaler){
            int gid = get_global_id(0);
            a_g[gid] = a_g[gid] / scaler;
        }

        __kernel void multiply(const unsigned int ac, const unsigned int bc, __global float * a, __global float * b, __global float * r) {
            int i = get_global_id(0); 
            r[i] = 0;
            for (unsigned int k = 0; k < ac; k++) {
                r[i] += a[k+((i/bc)*ac)] * b[(k*bc) + (i%bc)];
            }
        }
        """).build()

        programs = {}
        programs['testing'] = prg.testing
        programs['add'] = prg.add
        programs['subtract'] = prg.sub
        programs['element_wise_multiply'] = prg.element_wise_multiply
        programs['add_scaler'] = prg.add_scaler
        programs['sub_scaler'] = prg.sub_scaler
        programs['scaler_sub_from'] = prg.scaler_sub_from
        programs['scale'] = prg.scale
        programs['descale'] = prg.descale
        programs['add_inplace'] = prg.add_inplace
        programs['subtract_inplace'] = prg.sub_inplace
        programs['element_wise_multiply_inplace'] = prg.element_wise_multiply_inplace
        programs['add_scaler_inplace'] = prg.add_scaler_inplace
        programs['sub_scaler_inplace'] = prg.sub_scaler_inplace
        programs['scaler_sub_from_inplace'] = prg.scaler_sub_from_inplace
        programs['scale_inplace'] = prg.scale_inplace
        programs['descale_inplace'] = prg.descale_inplace
        programs['multiply'] = prg.multiply

        
        baseProgramNames = list(programs.keys())
        return cls(programs=programs, ctx=ctx, queue=queue, baseProgramNames=baseProgramNames)
    

    def loadCustomFunction(self, func_name, func):
        if func_name in self.programs:
            return
        
        func=func.replace("$i", "in[i]")

        prg = cl.Program(self.ctx, """
        __kernel void custom_func(
            __global float *in)
        {
            int i = get_global_id(0);
        """+func+"""    
        }
        """).build()

        self.programs[func_name] = prg.custom_func
        return func_name


    def unloadCustomFunction(self, func_name):
        if func_name in self.baseProgramNames:
            raise ValueError(func_name + " is already a standard program that cannot be unloaded")
        del self.programs[func_name]

    def testing(self, a, b):
        mf = cl.mem_flags

        a_g = cl.Buffer(self.ctx, mf.READ_WRITE | mf.COPY_HOST_PTR, hostbuf=a)
        b_g = cl.Buffer(self.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=b)

        self.programs['testing'](self.queue, a.shape, None, np.int32(len(a)), a_g, b_g)
        cl.enqueue_copy(self.queue, a, a_g)

    def _addInPlace(self, a, b):

        mf = cl.mem_flags
        a_g = cl.Buffer(self.ctx, mf.READ_WRITE | mf.COPY_HOST_PTR, hostbuf=a)
        b_g = cl.Buffer(self.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=b)

        self.programs['add_inplace'](self.queue, a.shape, None, a_g, b_g)
        cl.enqueue_copy(self.queue, a, a_g)

    def _subInPlace(self, a, b):

        mf = cl.mem_flags
        a_g = cl.Buffer(self.ctx, mf.READ_WRITE | mf.COPY_HOST_PTR, hostbuf=a)
        b_g = cl.Buffer(self.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=b)

        self.programs['subtract_inplace'](self.queue, a.shape, None, a_g, b_g)
        cl.enqueue_copy(self.queue, a, a_g)

    def _elementWiseMultiplyInPlace(self, a, b):

        mf = cl.mem_flags
        a_g = cl.Buffer(self.ctx, mf.READ_WRITE | mf.COPY_HOST_PTR, hostbuf=a)
        b_g = cl.Buffer(self.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=b)

        self.programs['element_wise_multiply_inplace'](self.queue, a.shape, None, a_g, b_g)
        cl.enqueue_copy(self.queue, a, a_g)

    def _addScalerInPlace(self, a, scaler):

        mf = cl.mem_flags
        a_g = cl.Buffer(self.ctx, mf.READ_WRITE | mf.COPY_HOST_PTR, hostbuf=a)

        self.programs['add_scaler_inplace'](self.queue, a.shape, None, a_g, np.float32(scaler))
        cl.enqueue_copy(self.queue, a, a_g)


    def _subScalerInPlace(self, a, scaler):

        mf = cl.mem_flags
        a_g = cl.Buffer(self.ctx, mf.READ_WRITE | mf.COPY_HOST_PTR, hostbuf=a)

        self.programs['sub_scaler_inplace'](self.queue, a.shape, None, a_g, np.float32(scaler))
        cl.enqueue_copy(self.queue, a, a_g)

    def _subScalerFromInPlace(self, a, scaler):

        mf = cl.mem_flags
        a_g = cl.Buffer(self.ctx, mf.READ_WRITE | mf.COPY_HOST_PTR, hostbuf=a)

        self.programs['scaler_sub_from_inplace'](self.queue, a.shape, None, a_g, np.float32(scaler))
        cl.enqueue_copy(self.queue, a, a_g)

    def _scaleInPlace(self, a, scaler):

        mf = cl.mem_flags
        a_g = cl.Buffer(self.ctx, mf.READ_WRITE | mf.COPY_HOST_PTR, hostbuf=a)

        self.programs['scale_inplace'](self.queue, a.shape, None, a_g, np.float32(scaler))
        cl.enqueue_copy(self.queue, a, a_g)

    def _descaleInPlace(self, a, scaler):

        mf = cl.mem_flags
        a_g = cl.Buffer(self.ctx, mf.READ_WRITE | mf.COPY_HOST_PTR, hostbuf=a)

        self.programs['descale_inplace'](self.queue, a.shape, None, a_g, np.float32(scaler))
        cl.enqueue_copy(self.queue, a, a_g)

    def _add(self, a, b):


        r_np=np.empty(len(a), dtype=a.dtype)

        mf = cl.mem_flags
        a_g = cl.Buffer(self.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=a)
        b_g = cl.Buffer(self.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=b)
        r_g = cl.Buffer(self.ctx, mf.WRITE_ONLY | mf.COPY_HOST_PTR, hostbuf=r_np)

        self.programs['add'](self.queue, a.shape, None, a_g, b_g, r_g)
        cl.enqueue_copy(self.queue, r_np, r_g)
        return r_np

    def _sub(self, a, b):

        r_np=np.empty(len(a), dtype=a.dtype)

        mf = cl.mem_flags
        a_g = cl.Buffer(self.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=a)
        b_g = cl.Buffer(self.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=b)
        r_g = cl.Buffer(self.ctx, mf.WRITE_ONLY | mf.COPY_HOST_PTR, hostbuf=r_np)

        self.programs['subtract'](self.queue, a.shape, None, a_g, b_g, r_g)
        cl.enqueue_copy(self.queue, r_np, r_g)
        return r_np

    def _elementWiseMultiply(self, a, b):

        r_np=np.empty(len(a), dtype=a.dtype)

        mf = cl.mem_flags
        a_g = cl.Buffer(self.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=a)
        b_g = cl.Buffer(self.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=b)
        r_g = cl.Buffer(self.ctx, mf.WRITE_ONLY | mf.COPY_HOST_PTR, hostbuf=r_np)

        self.programs['element_wise_multiply'](self.queue, a.shape, None, a_g, b_g, r_g)
        cl.enqueue_copy(self.queue, r_np, r_g)
        return r_np

    def _addScaler(self, a, scaler):

        r_np=np.empty(len(a), dtype=a.dtype)

        mf = cl.mem_flags
        a_g = cl.Buffer(self.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=a)
        r_g = cl.Buffer(self.ctx, mf.WRITE_ONLY | mf.COPY_HOST_PTR, hostbuf=r_np)

        self.programs['add_scaler'](self.queue, a.shape, None, a_g, np.float32(scaler), r_g)
        cl.enqueue_copy(self.queue, r_np, r_g)
        return r_np

    def _subScaler(self, a, scaler):

        r_np=np.empty(len(a), dtype=a.dtype)

        mf = cl.mem_flags
        a_g = cl.Buffer(self.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=a)
        r_g = cl.Buffer(self.ctx, mf.WRITE_ONLY | mf.COPY_HOST_PTR, hostbuf=r_np)

        self.programs['sub_scaler'](self.queue, a.shape, None, a_g, np.float32(scaler), r_g)
        cl.enqueue_copy(self.queue, r_np, r_g)
        return r_np

    def _subScalerFrom(self, a, scaler):

        r_np=np.empty(len(a), dtype=a.dtype)

        mf = cl.mem_flags
        a_g = cl.Buffer(self.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=a)
        r_g = cl.Buffer(self.ctx, mf.WRITE_ONLY | mf.COPY_HOST_PTR, hostbuf=r_np)

        self.programs['scaler_sub_from'](self.queue, a.shape, None, a_g, np.float32(scaler), r_g)
        cl.enqueue_copy(self.queue, r_np, r_g)
        return r_np


    def _scale(self, a, scaler):

        r_np=np.empty(len(a), dtype=a.dtype)

        mf = cl.mem_flags
        a_g = cl.Buffer(self.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=a)
        r_g = cl.Buffer(self.ctx, mf.WRITE_ONLY | mf.COPY_HOST_PTR, hostbuf=r_np)

        self.programs['scale'](self.queue, a.shape, None, a_g, np.float32(scaler), r_g)
        cl.enqueue_copy(self.queue, r_np, r_g)
        return r_np

    def _descale(self, a, scaler):

        r_np=np.empty(len(a), dtype=a.dtype)

        mf = cl.mem_flags
        a_g = cl.Buffer(self.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=a)
        r_g = cl.Buffer(self.ctx, mf.WRITE_ONLY | mf.COPY_HOST_PTR, hostbuf=r_np)

        self.programs['descale'](self.queue, a.shape, None, a_g, np.float32(scaler), r_g)
        cl.enqueue_copy(self.queue, r_np, r_g)
        return r_np

    def _multiply(self, a, b, a_rows:int, a_cols:int, b_rows:int, b_cols:int):

        res_np = np.empty(int(a_rows * b_cols), dtype=a.dtype)

        mf = cl.mem_flags
        a_g = cl.Buffer(self.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=a)
        b_g = cl.Buffer(self.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=b)
        res_g = cl.Buffer(self.ctx, mf.WRITE_ONLY, res_np.nbytes)

        self.programs['multiply'](self.queue, res_np.shape, None, np.int32(a_cols), np.int32(b_cols), a_g, b_g, res_g)

        cl.enqueue_copy(self.queue, res_np, res_g)
        return res_np

    def _sum(self, a):
        return a.sum()

    def _applyCustomFunctionInPlace(self, a, func_name):

        mf = cl.mem_flags
        a_g = cl.Buffer(self.ctx, mf.READ_WRITE | mf.COPY_HOST_PTR, hostbuf=a)

        self.programs[func_name](self.queue, a.shape, None, a_g)

        cl.enqueue_copy(self.queue, a, a_g)

    def _applyCustomFunction(self, a, func_name):

        r_np=np.empty(len(a), dtype=a.dtype)

        mf = cl.mem_flags
        a_g = cl.Buffer(self.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=a)
        r_g = cl.Buffer(self.ctx, mf.WRITE_ONLY | mf.COPY_HOST_PTR, hostbuf=r_np)

        self.programs[func_name](self.queue, a.shape, None, a_g, r_g)
        cl.enqueue_copy(self.queue, r_np, r_g)
        return r_np


from abc import ABC, abstractmethod

class LinearCalculations(ABC):


    def __init__(self):
        pass
    
    @classmethod
    @abstractmethod
    def getLinearCalculator(cls):
        pass

    @abstractmethod
    def loadCustomFunction(self, func_name, func):
        pass

    @abstractmethod
    def unloadCustomFunction(self, func_name):
        pass

    @abstractmethod
    def _addInPlace(self, a, b):
        pass

    @abstractmethod
    def _subInPlace(self, a, b):
        pass

    @abstractmethod
    def _elementWiseMultiplyInPlace(self, a, b):
        pass

    @abstractmethod
    def _addScalerInPlace(self, a, scaler):
        pass

    @abstractmethod
    def _subScalerInPlace(self, a, scaler):
        pass

    @abstractmethod
    def _subScalerFromInPlace(self, a, scaler):
        pass

    @abstractmethod
    def _scaleInPlace(self, a, scaler):
        pass

    @abstractmethod
    def _descaleInPlace(self, a, scaler):
        pass

    @abstractmethod
    def _add(self, a, b):
        pass

    @abstractmethod
    def _sub(self, a, b):
        pass

    @abstractmethod
    def _elementWiseMultiply(self, a, b):
        pass

    @abstractmethod
    def _addScaler(self, a, scaler):
        pass

    @abstractmethod
    def _subScaler(self, a, scaler):
        pass

    @abstractmethod
    def _subScalerFrom(self, a, scaler):
        pass

    @abstractmethod
    def _scale(self, a, scaler):
        pass

    @abstractmethod
    def _descale(self, a, scaler):
        pass

    @abstractmethod
    def _multiply(self, a, b, a_rows:int, a_cols:int, b_rows:int, b_cols:int):
        pass

    @abstractmethod
    def _applyCustomFunctionInPlace(self, a, func_name):
        pass

    @abstractmethod
    def _applyCustomFunction(self, a, func_name):
        pass

    @abstractmethod
    def _sum(self, a):
        pass


from typing import Any
import numpy as np
from parallellinear.calculations.CalculationsManager import CalculationsManager
from parallellinear.calculations.NumpyLinear import NumpyLinear




class Matrix:

    CALCULATIONS_MANAGER=CalculationsManager(NumpyLinear.getLinearCalculator(), np.float32)

    def __init__(self, rows:int, data:np.ndarray, calcManager=None):
        self.rows = rows
        self.columns=int(len(data)/rows)
        self.data=data
        if calcManager==None:
            self.calcManager = Matrix.CALCULATIONS_MANAGER
        else:
            self.calcManager = calcManager

    @classmethod
    def setCalculationsManager(cls, calculationsManager:CalculationsManager):
        Matrix.CALCULATIONS_MANAGER = calculationsManager

    @classmethod
    def loadCustomFunctionToCalculationsManager(cls, function_name:str, func:str):
        Matrix.CALCULATIONS_MANAGER.getCalculator().loadCustomFunction(function_name, func)

    @classmethod
    def random(cls, rows:int, columns:int, calcManager=None, random_low=0, random_high=1):
        if calcManager==None:
            calcManager = Matrix.CALCULATIONS_MANAGER
        out = cls(rows=rows, data=np.random.rand(rows*columns).astype(calcManager.getPrecision()), calcManager=calcManager)
        if random_low != 0 or random_high != 1:
            out.scale(random_high-random_low)
            out.addScaler(random_low)
        return out 

    @classmethod
    def fromFlatListGivenRowNumber(cls, rows:int, data:list, calcManager=None,):
        if calcManager==None:
            calcManager = Matrix.CALCULATIONS_MANAGER
        if len(data) % rows != 0:
                raise ValueError("Matrix from list requires the following assertion to be true len(second parameter) % first parameter == 0")
        return cls(rows=rows, data=np.array(data, dtype=calcManager.getPrecision()), calcManager=calcManager)

    
    @classmethod
    def zeros(cls, rows:int, columns:int, calcManager=None):
        if calcManager==None:
            calcManager = Matrix.CALCULATIONS_MANAGER
        return cls(rows=rows, data=np.zeros(rows*columns, dtype=calcManager.getPrecision()), calcManager=calcManager)

    @classmethod
    def filledWithValue(cls, rows:int, columns:int, value, calcManager=None):
        if calcManager==None:
            calcManager = Matrix.CALCULATIONS_MANAGER
        return cls(rows=rows, data=np.empty(rows*columns, dtype=calcManager.getPrecision()).fill(value), calcManager=calcManager)


    def getNumberOfColumns(self):
        return self.columns

    def getNumberOfRows(self):
        return self.rows

    def setAtPos(self, x, y, val):
        self.data[x*self.columns+y] = val
    
    def getAtPos(self, x, y):
        return self.data[x*self.columns+y]

    def __str__(self):
        return str(self.data)

    def __len__(self) -> int:
        return len(self.data)

    def getData(self):
        return self.data

    def exportToListRowNoAtLast(self):
        out=self.exportToList()
        out.append(self.rows)
        return out
    
    def exportToList(self):
        return self.data.tolist()


    def add(self, a, in_place = True) -> Any:
        if in_place:
            self.calcManager.getCalculator()._addInPlace(self.data, a.getData())
        else:
            return Matrix(self.rows, self.calcManager.getCalculator()._add(self.data, a.getData()), self.calcManager)

    def sub(self, a, in_place = True) -> Any:
        if in_place:
            self.calcManager.getCalculator()._subInPlace(self.data, a.getData())
        else:
            return Matrix(self.rows, self.calcManager.getCalculator()._sub(self, a), self.calcManager)

    def addScaler(self, scaler, in_place = True) -> Any:
        if in_place:
            self.calcManager.getCalculator()._addScalerInPlace(self.data, scaler)
        else:
            return Matrix(self.rows, self.calcManager.getCalculator()._addScaler(self.data, scaler), self.calcManager)

    def subScaler(self, scaler, in_place = True) -> Any:
        if in_place:
            self.calcManager.getCalculator()._subScalerInPlace(self.data, scaler)
        else:
            return Matrix(self.rows, self.calcManager.getCalculator()._subScaler(self.data, scaler), self.calcManager)

    def subScalerFrom(self, scaler, in_place = True) -> Any:
        if in_place:
            self.calcManager.getCalculator()._subScalerFromInPlace(self.data, scaler)
        else:
            return Matrix(self.rows, self.calcManager.getCalculator()._subScalerFrom(self.data, scaler), self.calcManager)

    def scale(self, scaler, in_place = True) -> Any:
        if in_place:
            self.calcManager.getCalculator()._scaleInPlace(self.data, scaler)
        else:
            return Matrix(self.rows, self.calcManager.getCalculator()._scale(self.data, scaler), self.calcManager)

    def descale(self, scaler, in_place = True) -> Any:
        if in_place:
            self.calcManager.getCalculator()._descale(self.data, scaler)
        else:
            return Matrix(self.rows, self.calcManager.getCalculator()._descale(self.data, scaler), self.calcManager)
    
    
    def __getitem__(self, indicies):
        return self.data[self.columns * indicies:self.columns*(indicies +1)]

    def transpose(self, in_place = True):
        
        bufferList = []
        for i in range(0, self.columns):
            for j in range(0, self.rows):
                bufferList.append(self.data[(j*self.columns)+i])
        

        if in_place:
            self.data = np.array(bufferList, dtype=self.calcManager.getPrecision())
            tmp=self.columns
            self.columns = self.rows
            self.rows = tmp
        else:
            return Matrix.fromFlatListGivenRowNumber(self.columns, bufferList)

    def multiply(self, a):
        return Matrix(self.rows, self.calcManager.getCalculator()._multiply(self.data, a.getData(), self.rows, self.columns, a.getNumberOfRows(), a.getNumberOfColumns()), self.calcManager)

    def applyCustomFunction(self, func_name, in_place = True):
        if in_place:
            self.calcManager.getCalculator()._applyCustomFunctionInPlace(self.data, func_name)
        else:
            return Matrix(self.rows, self.calcManager.getCalculator()._applyCustomFunction(self.data, func_name), self.calcManager)

    def sum(self):
        return self.calcManager.getCalculator()._sum(self.data)

    def elementWiseMultiply(self, a, in_place = True):
        if in_place:
            self.calcManager.getCalculator()._elementWiseMultiplyInPlace(self.data, a.getData())
        else:
            return Matrix(self.rows, self.calcManager.getCalculator()._elementWiseMultiply(self.data, a.getData()), self.calcManager)
            
        








from parallellinear.calculations.LinearCalculations import LinearCalculations
import numpy as np


class NumpyLinear(LinearCalculations):



    def __init__(self):
        self.customFunctions = {}

    @classmethod
    def getLinearCalculator(cls):
        return cls()
    
    # $i = 1/(1+exp(-$i));

    def loadCustomFunction(self, func_name, func):
        if func_name in self.customFunctions:
            return
        # .replace("$i = ", "def f(x): return ")
        interpolatedFunction=func.replace("$i = ", "").replace("$i", "a").replace("exp(", "np.exp(").removesuffix(";")
        self.customFunctions[func_name] = interpolatedFunction
        

    
    def unloadCustomFunction(self, func_name):
        del self.customFunctions[func_name]

    
    def _addInPlace(self, a, b):
        a += b

    
    def _subInPlace(self, a, b):
        a -= b

    
    def _elementWiseMultiplyInPlace(self, a, b):
        a *= b

    
    def _addScalerInPlace(self, a, scaler):
        a += scaler

    
    def _subScalerInPlace(self, a, scaler):
        a -= scaler

    
    def _subScalerFromInPlace(self, a, scaler):
        a = scaler - a

    
    def _scaleInPlace(self, a, scaler):
        a *= scaler

    
    def _descaleInPlace(self, a, scaler):
        a /= scaler

    
    def _add(self, a, b):
        return a + b

    
    def _sub(self, a, b):
        return a - b

    
    def _elementWiseMultiply(self, a, b):
        return a * b

    
    def _addScaler(self, a, scaler):
        return a + scaler

    
    def _subScaler(self, a, scaler):
        return a - scaler

    
    def _subScalerFrom(self, a, scaler):
        return scaler - a

    
    def _scale(self, a, scaler):
        return a * scaler

    
    def _descale(self, a, scaler):
        return a / scaler

    def _sum(self, a):
        return a.sum()

    
    def _multiply(self, a, b, a_rows:int, a_cols:int, b_rows:int, b_cols:int):
        amd=np.array([a[i:i+a_cols] for i in range(0, len(a), a_cols)], dtype=a.dtype)
        bmd=np.array([b[i:i+b_cols] for i in range(0, len(b), b_cols)], dtype=a.dtype).data
        return np.array(amd.dot(bmd).flat, dtype=a.dtype)

    
    def _applyCustomFunctionInPlace(self, a, func_name):
        a.data = eval(self.customFunctions[func_name]).data
    
    def _applyCustomFunction(self, a, func_name):
        return eval(self.customFunctions[func_name])


from typing import Any
from parallellinear.calculations.LinearCalculations import LinearCalculations
import numpy as np
from parallellinear.datatypes.Matrix import Matrix



class Vector(Matrix):

    def __init__(self, data, calcManager=None):
        super().__init__(1, data, calcManager)
           

    @classmethod
    def random(cls, length:int, calcManager = Matrix.CALCULATIONS_MANAGER, random_low=0, random_high=1):
        if calcManager==None:
            calcManager = Matrix.CALCULATIONS_MANAGER
        out = cls(data=np.random.rand(length).astype(calcManager.getPrecision()), calcManager=calcManager)
        if random_low != 0 or random_high != 1:
            out.scale(random_high-random_low)
            out.addScaler(random_low)
        return out 

    @classmethod
    def fromList(cls, data:list, calcManager = Matrix.CALCULATIONS_MANAGER):
        if calcManager==None:
            calcManager = Matrix.CALCULATIONS_MANAGER
        return cls(data=np.array(data, dtype=calcManager.getPrecision()), calcManager=calcManager)

    
    @classmethod
    def zeros(cls, length:int, calcManager = Matrix.CALCULATIONS_MANAGER):
        if calcManager==None:
            calcManager = Matrix.CALCULATIONS_MANAGER
        return cls(data=np.zeros(length, dtype=calcManager.getPrecision()), calcManager=calcManager)

    @classmethod
    def filledWithValue(cls, length:int, value, calcManager = Matrix.CALCULATIONS_MANAGER):
        return cls(data=np.empty(length, dtype=calcManager.getPrecision()).fill(value), calcManager=calcManager)



    def setAtPos(self, x, val):
        self.data[x] = val
    
    def getAtPos(self, x):
        return self.data[x]

    def __str__(self):
        return str(self.data)

    def exportToList(self):
        return self.data.tolist()

    def getData(self):
        return self.data

    def __len__(self) -> int:
        return len(self.data)

    def __getitem__(self, indicies):
        return self.data[indicies]


    def add(self, a, in_place = True) -> Any:
        if in_place:
            self.calcManager.getCalculator()._addInPlace(self.data, a.getData())
        else:
            return Vector(self.calcManager.getCalculator()._add(self.data, a.getData()), self.calcManager)

    def sub(self, a, in_place = True) -> Any:
        if in_place:
            self.calcManager.getCalculator()._subInPlace(self.data, a.getData())
        else:
            return Vector(self.calcManager.getCalculator()._sub(self.data, a.data), self.calcManager)

    def addScaler(self, scaler, in_place = True) -> Any:
        if in_place:
            self.calcManager.getCalculator()._addScalerInPlace(self.data, scaler)
        else:
            return Vector(self.calcManager.getCalculator()._addScaler(self.data, scaler), self.calcManager)

    def subScaler(self, scaler, in_place = True) -> Any:
        if in_place:
            self.calcManager.getCalculator()._subScalerInPlace(self.data, scaler)
        else:
            return Vector(self.calcManager.getCalculator()._subScaler(self.data, scaler), self.calcManager)

    def subScalerFrom(self, scaler, in_place = True) -> Any:
        if in_place:
            self.calcManager.getCalculator()._subScalerFromInPlace(self.data, scaler)
        else:
            return Vector(self.calcManager.getCalculator()._subScalerFrom(self.data, scaler), self.calcManager)

    def scale(self, scaler, in_place = True) -> Any:
        if in_place:
            self.calcManager.getCalculator()._scaleInPlace(self.data, scaler)
        else:
            return Vector(self.calcManager.getCalculator()._scale(self.data, scaler), self.calcManager)

    def descale(self, scaler, in_place = True) -> Any:
        if in_place:
            self.calcManager.getCalculator()._descaleInPlace(self.data, scaler)
        else:
            return Vector(self.calcManager.getCalculator()._descale(self.data, scaler), self.calcManager)

    def applyCustomFunction(self, func_name, in_place = True):
        if in_place:
            self.calcManager.getCalculator()._applyCustomFunctionInPlace(self.data, func_name)
        else:
            return Vector(self.calcManager.getCalculator()._applyCustomFunction(self.data, func_name), self.calcManager)

    def magnitude(a):
        return np.sqrt(np.sum((lambda x:np.square(x))(a)))

    def elementWiseMultiply(self, a, in_place = True):
        if in_place:
            self.calcManager.getCalculator()._elementWiseMultiplyInPlace(self.data, a.getData())
        else:
            return Vector(self.calcManager.getCalculator()._elementWiseMultiply(self.data, a.getData()), self.calcManager)


from parallellinear.calculations.LinearCalculations import LinearCalculations
import numpy as np


class CalculationsManager:

    def __init__(self, calculator:LinearCalculations, precision:np.dtype):
        self.calculator = calculator
        self.precision = precision

    def getPrecision(self):
        return self.precision

    def getCalculator(self):
        return self.calculator